#include <iostream>
#include "toolbox.hpp"
#include "format_time.hpp"

/**
 * @brief Test if a given result is equal of the expected one and log result
 *
 * @tparam T type of returning values
 * @param name of the unit test
 * @param expected result of the function call
 * @param result of the function
 */
template<typename T>
void Assert(const char* const name, const T& expected, const T& result) noexcept {
	if (expected != result)
		std::cerr << AnsiColor::Red << "[ ] - " << name << AnsiColor::Reset << " - Expected '" << expected << "' but got '" << result << "' instead\n";
	else
		std::cout << AnsiColor::Green << "[✓] - " << name << AnsiColor::Reset << "\n";
}

/**
 * @brief Test suite for the format_time output
 */
void format_time_test(void) noexcept {
	std::cout << AnsiColor::BoldWhite << "\tTesting format_time str suite" << AnsiColor::Reset << "\n";

	// https://en.wikipedia.org/wiki/Unit_of_time
	Assert("format_time str null",                      std::string("0s"),                                    format_time(u64(0)));
	Assert("format_time str second",                    std::string("1s"),                                    format_time(u64(1)));
	Assert("format_time str decasecond",                std::string("10s"),                                   format_time(u64(10)));
	Assert("format_time str minute",                    std::string("1m"),                                    format_time(u64(60)));
	Assert("format_time str milliday",                  std::string("1m 26s"),                                format_time(u64(86))); // missing 0.4s due to precision
	Assert("format_time str hectosecond",               std::string("1m 40s"),                                format_time(u64(100)));
	Assert("format_time str kilosecond",                std::string("16m 40s"),                               format_time(u64(1e3)));
	Assert("format_time str hour",                      std::string("1h"),                                    format_time(u64(3600)));
	Assert("format_time str day",                       std::string("1j"),                                    format_time(u64(86400)));
	Assert("format_time str week/sennight",             std::string("1w"),                                    format_time(u64(604800)));
	Assert("format_time str megasecond",                std::string("1w 4j 13h 46m 40s"),                     format_time(u64(1e6)));
	Assert("format_time str fortnight",                 std::string("2w"),                                    format_time(u64(1209600)));
	Assert("format_time str lunar month (draconitic)",  std::string("3w 6j 5h 5m 35s"),                       format_time(u64(2351135))); // missing 0.8 due to precision
	Assert("format_time str lunar month (tropical)",    std::string("3w 6j 7h 43m 4s"),                       format_time(u64(2360584))); // missing 0.7 due to precision
	Assert("format_time str lunar month (sidereal)",    std::string("3w 6j 7h 43m 11s"),                      format_time(u64(2360591))); // missing 0.6 to precision
	Assert("format_time str lunar month (anomalistic)", std::string("3w 6j 13h 18m 33s"),                     format_time(u64(2380713))); // missing 0.2 due to precision
	Assert("format_time str lunar month (synodic)",     std::string("4w 1j 12h 44m 2s"),                      format_time(u64(2551442))); // missing 0.9 due to precision
	Assert("format_time str month",                     std::string("1M"),                                    format_time(u64(2629800)));
	Assert("format_time str quarantine",                std::string("1M 1w 2j 13h 30m"),                      format_time(u64(3456e3)));
	Assert("format_time str semester",                  std::string("4M 4j 6h"),                              format_time(u64(10886400)));
	Assert("format_time str lunar year",                std::string("11M 2w 5j 13h 22m 48s"),                 format_time(u64(30617568)));
	Assert("format_time str common year",               std::string("11M 4w 2j 4h 30m"),                      format_time(u64(31536e3)));
	Assert("format_time str year",                      std::string("1y"),                                    format_time(u64(31557600)));
	Assert("format_time str tropical year",             std::string("11M 4w 2j 10h 18m 45s"),                 format_time(u64(31556925))); // missing 0.216 due to precision
	Assert("format_time str gregorian year",            std::string("11M 4w 2j 10h 19m 12s"),                 format_time(u64(31556952)));
	Assert("format_time str sidereal year",             std::string("1y 9m 9s"),                              format_time(u64(31558149))); // missing 0.7635456 due to precision
	Assert("format_time str leap year",                 std::string("1y 18h"),                                format_time(u64(31622400)));
	Assert("format_time str olympiad",                  std::string("4y"),                                    format_time(u64(126230400)));
	Assert("format_time str lusturm",                   std::string("5y"),                                    format_time(u64(157788e3)));
	Assert("format_time str decade",                    std::string("10y"),                                   format_time(u64(315576e3)));
	Assert("format_time str indiction",                 std::string("15y"),                                   format_time(u64(473364e3)));
	Assert("format_time str score",                     std::string("20y"),                                   format_time(u64(631152e3)));
	Assert("format_time str gigasecond",                std::string("31y 8M 1w 19h 46m 40s"),                 format_time(u64(1e9)));
	Assert("format_time str jubilee",                   std::string("50y"),                                   format_time(u64(157788e4)));
	Assert("format_time str century",                   std::string("1c"),                                    format_time(u64(315576e4)));

	Assert("format_time str millennium",                std::string("10c"),                                   format_time(u64(315576e5)));
	Assert("format_time str age",                       std::string("257c 72y"),                              format_time(u64(813302467200)));
	Assert("format_time str terasecond",                std::string("3168c 80y 10M 2w 2j 8h 46m 40s"),        format_time(u64(1e13)));
	Assert("format_time str megaannum",                 std::string("10000c"),                                format_time(u64(315576e8)));
	Assert("format_time str petasecond",                std::string("316880c 87y 9M 3w 2j 9h 16m 40s"),       format_time(u64(1e15)));
	Assert("format_time str galactic year",             std::string("2300000c"),                              format_time(u64(7258248e9)));
	Assert("format_time str kalpa",                     std::string("43200000c"),                             format_time(u64(136328832e9)));
	Assert("format_time str eon",                       std::string("10000000c"),                             format_time(u64(315576e11)));
	Assert("format_time str exasecond",                 std::string("316880878c 14y 1w 3j 13h 46m 40s"),      format_time(u64(1e18)));
	// Cannot use number bigger than currently supported ISO C++
	//Assert("format_time str zettasecond",               std::string(""),                                      format_time(u64(1e21)));
	//Assert("format_time str yottasecond",               std::string(""),                                      format_time(u64(1e24)));
	//Assert("format_time str ronnasecond",               std::string(""),                                      format_time(u64(1e27)));
	//Assert("format_time str quettasecond",              std::string(""),                                      format_time(u64(1e30)));
	// uint64_t MAX == 2**64 == 18446744073709551615 == -1
	Assert("format_time str max",                       std::string("5845420460c 90y 7M 2w 1j 17h 30m 15s"),  format_time(u64(-1)));
	Assert("format_time str max",                       std::string("5845420460c 90y 7M 2w 1j 17h 30m 15s"),  format_time(18446744073709551615ull));
	Assert("format_time str longest",                   std::string("1000000000c 10y 10M 3w 6j 10h 10m 10s"), format_time(3155760000344243410ull));
}

/**
 * @brief Test suite for the format_time_ns output
 */
void format_time_ns_test(void) noexcept {
	std::cout << AnsiColor::BoldWhite << "\tTesting format_time_ns str suite" << AnsiColor::Reset << "\n";

	// https://en.wikipedia.org/wiki/Unit_of_time
	Assert("format_time_ns str null",                      std::string("0ns"),                                            format_time_ns(u64(0)));
	Assert("format_time_ns str nanosecond",                std::string("1ns"),                                            format_time_ns(u64(1)));
	Assert("format_time_ns str shake",                     std::string("10ns"),                                           format_time_ns(u64(10)));
	Assert("format_time_ns str microsecond",               std::string("1us"),                                            format_time_ns(u64(1e3)));
	Assert("format_time_ns str millisecond",               std::string("1ms"),                                            format_time_ns(u64(1e6)));
	Assert("format_time_ns str centisecond",               std::string("10ms"),                                           format_time_ns(u64(1e7)));
	Assert("format_time_ns str decisecond",                std::string("100ms"),                                          format_time_ns(u64(1e8)));
	Assert("format_time_ns str second",                    std::string("1s"),                                             format_time_ns(u64(1e9)));
	Assert("format_time_ns str decasecond",                std::string("10s"),                                            format_time_ns(u64(1e10)));
	Assert("format_time_ns str minute",                    std::string("1m"),                                             format_time_ns(u64(6e10)));
	Assert("format_time_ns str milliday",                  std::string("1m 26s 400ms"),                                   format_time_ns(u64(864e8)));
	Assert("format_time_ns str hectosecond",               std::string("1m 40s"),                                         format_time_ns(u64(1e11)));
	Assert("format_time_ns str kilosecond",                std::string("16m 40s"),                                        format_time_ns(u64(1e12)));
	Assert("format_time_ns str hour",                      std::string("1h"),                                             format_time_ns(u64(36e11)));
	Assert("format_time_ns str day",                       std::string("1j"),                                             format_time_ns(u64(864e11)));
	Assert("format_time_ns str week/sennight",             std::string("1w"),                                             format_time_ns(u64(6048e11)));
	Assert("format_time_ns str megasecond",                std::string("1w 4j 13h 46m 40s"),                              format_time_ns(u64(1e15)));
	Assert("format_time_ns str fortnight",                 std::string("2w"),                                             format_time_ns(u64(12096e11)));
	Assert("format_time_ns str lunar month (draconitic)",  std::string("3w 6j 5h 5m 35s 800ms"),                          format_time_ns(u64(23511358e8)));
	Assert("format_time_ns str lunar month (tropical)",    std::string("3w 6j 7h 43m 4s 700ms"),                          format_time_ns(u64(23605847e8)));
	Assert("format_time_ns str lunar month (sidereal)",    std::string("3w 6j 7h 43m 11s 600ms"),                         format_time_ns(u64(23605916e8)));
	Assert("format_time_ns str lunar month (anomalistic)", std::string("3w 6j 13h 18m 33s 200ms"),                        format_time_ns(u64(23807132e8)));
	Assert("format_time_ns str lunar month (synodic)",     std::string("4w 1j 12h 44m 2s 900ms"),                         format_time_ns(u64(25514429e8)));
	Assert("format_time_ns str month",                     std::string("1M"),                                             format_time_ns(u64(26298e11)));
	Assert("format_time_ns str quarantine",                std::string("1M 1w 2j 13h 30m"),                               format_time_ns(u64(3456e12)));
	Assert("format_time_ns str semester",                  std::string("4M 4j 6h"),                                       format_time_ns(u64(108864e11)));
	Assert("format_time_ns str lunar year",                std::string("11M 2w 5j 13h 22m 48s"),                          format_time_ns(u64(30617568e9)));
	Assert("format_time_ns str common year",               std::string("11M 4w 2j 4h 30m"),                               format_time_ns(u64(31536e12)));
	Assert("format_time_ns str year",                      std::string("1y"),                                             format_time_ns(u64(315576e11)));
	Assert("format_time_ns str tropical year",             std::string("11M 4w 2j 10h 18m 45s"),                          format_time_ns(u64(31556925e9)));
	Assert("format_time_ns str gregorian year",            std::string("11M 4w 2j 10h 19m 12s"),                          format_time_ns(u64(31556952e9)));
	Assert("format_time_ns str sidereal year",             std::string("1y 9m 9s"),                                       format_time_ns(u64(31558149e9)));
	Assert("format_time_ns str leap year",                 std::string("1y 18h"),                                         format_time_ns(u64(316224e11)));
	Assert("format_time_ns str olympiad",                  std::string("4y"),                                             format_time_ns(u64(1262304e11)));
	Assert("format_time_ns str lusturm",                   std::string("5y"),                                             format_time_ns(u64(157788e12)));
	Assert("format_time_ns str decade",                    std::string("10y"),                                            format_time_ns(u64(315576e12)));
	Assert("format_time_ns str indiction",                 std::string("15y"),                                            format_time_ns(u64(473364e12)));
	Assert("format_time_ns str score",                     std::string("20y"),                                            format_time_ns(u64(631152e12)));
	Assert("format_time_ns str gigasecond",                std::string("31y 8M 1w 19h 46m 40s"),                          format_time_ns(u64(1e18)));
	Assert("format_time_ns str jubilee",                   std::string("50y"),                                            format_time_ns(u64(157788e13)));
	Assert("format_time_ns str century",                   std::string("1c"),                                             format_time_ns(u64(315576e13)));
	// Cannot use number bigger than currently supported ISO C++
	//Assert("format_time_ns str millennium",                std::string("10c"),                                            format_time_ns(u64(315576e14)));
	//Assert("format_time_ns str age",                       std::string("257c 72y"),                                       format_time_ns(u64(8133024672e11)));
	//Assert("format_time_ns str terasecond",                std::string("3168c 80y 10M 2w 2j 8h 46m 40s"),                 format_time_ns(u64(1e22)));
	//Assert("format_time_ns str megaannum",                 std::string("10000c"),                                         format_time_ns(u64(315576e17)));
	//Assert("format_time_ns str petasecond",                std::string("316880c 87y 9M 3w 2j 9h 16m 40s"),                format_time_ns(u64(1e24)));
	//Assert("format_time_ns str galactic year",             std::string("2300000c"),                                       format_time_ns(u64(7258248e18)));
	//Assert("format_time_ns str eon",                       std::string("10000000c"),                                      format_time_ns(u64(136328832e18)));
	//Assert("format_time_ns str kalpa",                     std::string("43200000c"),                                      format_time_ns(u64(315576e20)));
	//Assert("format_time_ns str exasecond",                 std::string("316880878c 14y 1w 3j 13h 46m 40s"),               format_time_ns(u64(1e27)));
	//Assert("format_time_ns str zettasecond",               std::string(""),                                               format_time_ns(u64(1e30)));
	//Assert("format_time_ns str yottasecond",               std::string(""),                                               format_time_ns(u64(1e33)));
	//Assert("format_time_ns str ronnasecond",               std::string(""),                                               format_time_ns(u64(1e36)));
	//Assert("format_time_ns str quettasecond",              std::string(""),                                               format_time_ns(u64(1e39)));
	// uint64_t_MAX == 2**64 == 18446744073709551615 == -1
	Assert("format_time_ns str max",                       std::string("5c 84y 6M 2w 1j 8h 34m 33s 709ms 551us 615ns"),   format_time_ns(u64(-1)));
	Assert("format_time_ns str max",                       std::string("5c 84y 6M 2w 1j 8h 34m 33s 709ms 551us 615ns"),   format_time_ns(18446744073709551615ull));
	Assert("format_time_ns str longest",                   std::string("1c 10y 10M 3w 6j 10h 10m 10s 100ms 100us 100ns"), format_time_ns(3500003410100100100ull));
}

int32_t main(void) noexcept {
	format_time_test();
	format_time_ns_test();
}
